@@ -1,3 +1,5 @@
+#    -*- encoding: utf-8 -*-
+#
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
 #    a copy of the License at
@@ -14,18 +16,20 @@
 
 import contextlib
 import itertools
+from unittest import mock
 
-import mock
 from oslotest import base as oslo_test_base
-import six
 import sqlalchemy as sqla
+from sqlalchemy.engine import url as sqla_url
 from sqlalchemy import event
+import sqlalchemy.exc
+from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.orm import mapper
 
 from oslo_db import exception
+from oslo_db.sqlalchemy import engines
 from oslo_db.sqlalchemy import exc_filters
-from oslo_db.sqlalchemy import session
-from oslo_db.sqlalchemy import test_base
+from oslo_db.tests.sqlalchemy import base as test_base
 from oslo_db.tests import utils as test_utils
 
 _TABLE_NAME = '__tmp__test__tmp__'
@@ -45,12 +49,23 @@ class _SQLAExceptionMatcher(object):
         if isinstance(exception_type, (type, tuple)):
             self.assertTrue(issubclass(exc.__class__, exception_type))
         else:
-            self.assertEqual(exc.__class__.__name__, exception_type)
-        self.assertEqual(str(exc.orig).lower(), message.lower())
+            self.assertEqual(exception_type, exc.__class__.__name__)
+        if isinstance(message, tuple):
+            self.assertEqual(
+                [m.lower() if isinstance(m, str) else m for m in message],
+                [a.lower() if isinstance(a, str) else a for a in exc.orig.args]
+            )
+        else:
+            self.assertEqual(message.lower(), str(exc.orig).lower())
         if sql is not None:
-            self.assertEqual(exc.statement, sql)
-        if params is not None:
-            self.assertEqual(exc.params, params)
+            if params is not None:
+                if '?' in exc.statement:
+                    self.assertEqual(sql, exc.statement)
+                    self.assertEqual(params, exc.params)
+                else:
+                    self.assertEqual(sql % params, exc.statement % exc.params)
+            else:
+                self.assertEqual(sql, exc.statement)
 
 
 class TestsExceptionFilter(_SQLAExceptionMatcher, oslo_test_base.BaseTestCase):
@@ -63,6 +78,9 @@ class TestsExceptionFilter(_SQLAExceptionMatcher, oslo_test_base.BaseTestCase):
 
         """
 
+    class DataError(Error):
+        pass
+
     class OperationalError(Error):
         pass
 
@@ -205,14 +223,14 @@ class TestFallthroughsAndNonDBAPI(TestsExceptionFilter):
         # or at least not documented anywhere.
         uee_ref = None
         try:
-            six.u('\u2435').encode('ascii')
+            '\u2435'.encode('ascii')
         except UnicodeEncodeError as uee:
             # Python3.x added new scoping rules here (sadly)
             # http://legacy.python.org/dev/peps/pep-3110/#semantic-changes
             uee_ref = uee
 
         self._run_test(
-            "postgresql", six.u('select \u2435'),
+            'postgresql', 'select \u2435',
             uee_ref,
             exception.DBInvalidUnicodeParameter
             )
@@ -226,7 +244,212 @@ class TestFallthroughsAndNonDBAPI(TestsExceptionFilter):
         self.assertEqual("mysqldb has an attribute error", matched.args[0])
 
 
-class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
+class TestNonExistentConstraint(
+        _SQLAExceptionMatcher,
+        test_base._DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentConstraint, self).setUp()
+
+        meta = sqla.MetaData(bind=self.engine)
+
+        self.table_1 = sqla.Table(
+            "resource_foo", meta,
+            sqla.Column("id", sqla.Integer, primary_key=True),
+            mysql_engine='InnoDB',
+            mysql_charset='utf8',
+        )
+        self.table_1.create()
+
+
+class TestNonExistentConstraintPostgreSQL(
+        TestNonExistentConstraint,
+        test_base._PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentConstraint,
+            self.engine.execute,
+            sqla.schema.DropConstraint(
+                sqla.ForeignKeyConstraint(["id"], ["baz.id"],
+                                          name="bar_fkey",
+                                          table=self.table_1)),
+        )
+        self.assertInnerException(
+            matched,
+            "ProgrammingError",
+            "constraint \"bar_fkey\" of relation "
+            "\"resource_foo\" does not exist\n",
+            "ALTER TABLE resource_foo DROP CONSTRAINT bar_fkey",
+        )
+        self.assertEqual("resource_foo", matched.table)
+        self.assertEqual("bar_fkey", matched.constraint)
+
+
+class TestNonExistentConstraintMySQL(
+        TestNonExistentConstraint,
+        test_base._MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentConstraint,
+            self.engine.execute,
+            sqla.schema.DropConstraint(
+                sqla.ForeignKeyConstraint(["id"], ["baz.id"],
+                                          name="bar_fkey",
+                                          table=self.table_1)),
+        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              (sqlalchemy.exc.InternalError,
+                               sqlalchemy.exc.OperationalError))
+        if matched.table is not None:
+            self.assertEqual("resource_foo", matched.table)
+        if matched.constraint is not None:
+            self.assertEqual("bar_fkey", matched.constraint)
+
+
+class TestNonExistentTable(
+        _SQLAExceptionMatcher,
+        test_base._DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentTable, self).setUp()
+
+        self.meta = sqla.MetaData(bind=self.engine)
+
+        self.table_1 = sqla.Table(
+            "foo", self.meta,
+            sqla.Column("id", sqla.Integer, primary_key=True),
+            mysql_engine='InnoDB',
+            mysql_charset='utf8',
+        )
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        self.assertInnerException(
+            matched,
+            "OperationalError",
+            "no such table: foo",
+            "\nDROP TABLE foo",
+        )
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentTablePostgreSQL(
+        TestNonExistentTable,
+        test_base._PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        self.assertInnerException(
+            matched,
+            "ProgrammingError",
+            "table \"foo\" does not exist\n",
+            "\nDROP TABLE foo",
+        )
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentTableMySQL(
+        TestNonExistentTable,
+        test_base._MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentTable,
+            self.engine.execute,
+            sqla.schema.DropTable(self.table_1),
+        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              (sqlalchemy.exc.InternalError,
+                               sqlalchemy.exc.OperationalError))
+        self.assertEqual("foo", matched.table)
+
+
+class TestNonExistentDatabase(
+        _SQLAExceptionMatcher,
+        test_base._DbTestCase):
+
+    def setUp(self):
+        super(TestNonExistentDatabase, self).setUp()
+
+        url = sqla_url.make_url(str(self.engine.url))
+
+        # TODO(zzzeek): remove hasattr() conditional in favor of "url.set()"
+        # when SQLAlchemy 1.4 is the minimum version in requirements
+        if hasattr(url, "set"):
+            self.url = url.set(database="non_existent_database")
+        else:
+            # TODO(zzzeek): remove when SQLAlchemy 1.4
+            # is the minimum version in requirements
+            url.database = 'non_existent_database'
+            self.url = url
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            sqla_url.make_url(
+                'sqlite:////non_existent_dir/non_existent_database')
+        )
+        self.assertIsNone(matched.database)
+        self.assertInnerException(
+            matched,
+            sqlalchemy.exc.OperationalError,
+            'unable to open database file',
+        )
+
+
+class TestNonExistentDatabaseMySQL(
+        TestNonExistentDatabase,
+        test_base._MySQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            self.url
+        )
+        self.assertEqual('non_existent_database', matched.database)
+        # NOTE(rpodolyaka) cannot check precisely with assertInnerException
+        # since MySQL errors are not the same depending on its version
+        self.assertIsInstance(
+            matched.inner_exception,
+            (sqlalchemy.exc.InternalError, sqlalchemy.exc.OperationalError),
+        )
+
+
+class TestNonExistentDatabasePostgreSQL(
+        TestNonExistentDatabase,
+        test_base._PostgreSQLOpportunisticTestCase):
+
+    def test_raise(self):
+        matched = self.assertRaises(
+            exception.DBNonExistentDatabase,
+            engines.create_engine,
+            self.url
+        )
+        self.assertEqual('non_existent_database', matched.database)
+        self.assertInnerException(
+            matched,
+            sqlalchemy.exc.OperationalError,
+            'fatal:  database "non_existent_database" does not exist\n',
+        )
+
+
+class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base._DbTestCase):
 
     def setUp(self):
         super(TestReferenceErrorSQLite, self).setUp()
@@ -300,7 +523,7 @@ class TestReferenceErrorSQLite(_SQLAExceptionMatcher, test_base.DbTestCase):
 
 
 class TestReferenceErrorPostgreSQL(TestReferenceErrorSQLite,
-                                   test_base.PostgreSQLOpportunisticTestCase):
+                                   test_base._PostgreSQLOpportunisticTestCase):
     def test_raise(self):
         params = {'id': 1, 'foo_id': 2}
         matched = self.assertRaises(
@@ -351,7 +574,7 @@ class TestReferenceErrorPostgreSQL(TestReferenceErrorSQLite,
 
 
 class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
-                              test_base.MySQLOpportunisticTestCase):
+                              test_base._MySQLOpportunisticTestCase):
     def test_raise(self):
         matched = self.assertRaises(
             exception.DBReferenceError,
@@ -359,16 +582,11 @@ class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
             self.table_2.insert({'id': 1, 'foo_id': 2})
         )
 
-        self.assertInnerException(
-            matched,
-            "IntegrityError",
-            "(1452, 'Cannot add or update a child row: a "
-            "foreign key constraint fails (`{0}`.`resource_entity`, "
-            "CONSTRAINT `foo_fkey` FOREIGN KEY (`foo_id`) REFERENCES "
-            "`resource_foo` (`id`))')".format(self.engine.url.database),
-            "INSERT INTO resource_entity (id, foo_id) VALUES (%s, %s)",
-            (1, 2)
-        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(matched.inner_exception.orig.args[0], 1452)
         self.assertEqual("resource_entity", matched.table)
         self.assertEqual("foo_fkey", matched.constraint)
         self.assertEqual("foo_id", matched.key)
@@ -387,16 +605,11 @@ class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
                 self.table_2.insert({'id': 1, 'foo_id': 2})
             )
 
-        self.assertInnerException(
-            matched,
-            "IntegrityError",
-            '(1452, \'Cannot add or update a child row: a '
-            'foreign key constraint fails ("{0}"."resource_entity", '
-            'CONSTRAINT "foo_fkey" FOREIGN KEY ("foo_id") REFERENCES '
-            '"resource_foo" ("id"))\')'.format(self.engine.url.database),
-            "INSERT INTO resource_entity (id, foo_id) VALUES (%s, %s)",
-            (1, 2)
-        )
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(matched.inner_exception.orig.args[0], 1452)
         self.assertEqual("resource_entity", matched.table)
         self.assertEqual("foo_fkey", matched.constraint)
         self.assertEqual("foo_id", matched.key)
@@ -411,24 +624,124 @@ class TestReferenceErrorMySQL(TestReferenceErrorSQLite,
             self.engine.execute,
             self.table_1.delete()
         )
-        self.assertInnerException(
-            matched,
-            "IntegrityError",
-            "(1451, 'cannot delete or update a parent row: a foreign key "
-            "constraint fails (`{0}`.`resource_entity`, "
-            "constraint `foo_fkey` "
-            "foreign key (`foo_id`) references "
-            "`resource_foo` (`id`))')".format(self.engine.url.database),
-            "DELETE FROM resource_foo",
-            (),
-        )
-
+        # NOTE(jd) Cannot check precisely with assertInnerException since MySQL
+        # error are not the same depending on its version…
+        self.assertIsInstance(matched.inner_exception,
+                              sqlalchemy.exc.IntegrityError)
+        self.assertEqual(1451, matched.inner_exception.orig.args[0])
         self.assertEqual("resource_entity", matched.table)
         self.assertEqual("foo_fkey", matched.constraint)
         self.assertEqual("foo_id", matched.key)
         self.assertEqual("resource_foo", matched.key_table)
 
 
+class TestExceptionCauseMySQLSavepoint(test_base._MySQLOpportunisticTestCase):
+    def setUp(self):
+        super(TestExceptionCauseMySQLSavepoint, self).setUp()
+
+        Base = declarative_base()
+
+        class A(Base):
+            __tablename__ = 'a'
+
+            id = sqla.Column(sqla.Integer, primary_key=True)
+
+            __table_args__ = {'mysql_engine': 'InnoDB'}
+
+        Base.metadata.create_all(self.engine)
+
+        self.A = A
+
+    def test_cause_for_failed_flush_plus_no_savepoint(self):
+        session = self.sessionmaker()
+
+        with session.begin():
+            session.add(self.A(id=1))
+        try:
+            with session.begin():
+                try:
+                    with session.begin_nested():
+                        session.execute("rollback")
+                        session.add(self.A(id=1))
+                # outermost is the failed SAVEPOINT rollback
+                # from the "with session.begin_nested()"
+                except exception.DBError as dbe_inner:
+                    # in SQLA 1.1+, the rollback() method of Session
+                    # catches the error and repairs the state of the
+                    # session even though the SAVEPOINT was lost;
+                    # the net result here is that one exception is thrown
+                    # instead of two.  This is SQLAlchemy ticket #3680
+                    self.assertTrue(
+                        isinstance(
+                            dbe_inner.cause,
+                            exception.DBDuplicateEntry
+                        )
+                    )
+        except exception.DBError as dbe_outer:
+            self.assertTrue(
+                isinstance(
+                    dbe_outer.cause,
+                    exception.DBDuplicateEntry
+                )
+            )
+
+        # resets itself afterwards
+        try:
+            with session.begin():
+                session.add(self.A(id=1))
+        except exception.DBError as dbe_outer:
+            self.assertIsNone(dbe_outer.cause)
+
+    def test_rollback_doesnt_interfere_with_killed_conn(self):
+        session = self.sessionmaker()
+
+        session.begin()
+        try:
+            session.execute("select 1")
+
+            # close underying DB connection
+            session.connection().connection.connection.close()
+
+            # alternate approach, but same idea:
+            # conn_id = session.scalar("select connection_id()")
+            # session.execute("kill connection %s" % conn_id)
+
+            # try using it, will raise an error
+            session.execute("select 1")
+        except exception.DBConnectionError:
+            # issue being tested is that this session.rollback()
+            # does not itself try to re-connect and raise another
+            # error.
+            session.rollback()
+        else:
+            assert False, "no exception raised"
+
+    def test_savepoint_rollback_doesnt_interfere_with_killed_conn(self):
+        session = self.sessionmaker()
+
+        session.begin()
+        try:
+            session.begin_nested()
+            session.execute("select 1")
+
+            # close underying DB connection
+            session.connection().connection.connection.close()
+
+            # alternate approach, but same idea:
+            # conn_id = session.scalar("select connection_id()")
+            # session.execute("kill connection %s" % conn_id)
+
+            # try using it, will raise an error
+            session.execute("select 1")
+        except exception.DBConnectionError:
+            # issue being tested is that this session.rollback()
+            # does not itself try to re-connect and raise another
+            # error.
+            session.rollback()
+        else:
+            assert False, "no exception raised"
+
+
 class TestConstraint(TestsExceptionFilter):
     def test_postgresql(self):
         matched = self._run_test(
@@ -483,11 +796,15 @@ class TestDuplicate(TestsExceptionFilter):
             "PRIMARY KEY must be unique 'insert into t values(10)'",
             expected_columns=[])
 
-    def test_mysql_mysqldb(self):
+    def test_mysql_pymysql(self):
         self._run_dupe_constraint_test(
             "mysql",
             '(1062, "Duplicate entry '
             '\'2-3\' for key \'uniq_tbl0a0b\'")', expected_value='2-3')
+        self._run_dupe_constraint_test(
+            "mysql",
+            '(1062, "Duplicate entry '
+            '\'\' for key \'uniq_tbl0a0b\'")', expected_value='')
 
     def test_mysql_mysqlconnector(self):
         self._run_dupe_constraint_test(
@@ -512,6 +829,14 @@ class TestDuplicate(TestsExceptionFilter):
             expected_value='2'
         )
 
+    def test_mysql_duplicate_entry_key_start_with_tablename(self):
+        self._run_dupe_constraint_test(
+            "mysql",
+            "1062 (23000): Duplicate entry '2' for key 'tbl.uniq_tbl0b'",
+            expected_columns=['b'],
+            expected_value='2'
+        )
+
     def test_mysql_binary(self):
         self._run_dupe_constraint_test(
             "mysql",
@@ -530,6 +855,24 @@ class TestDuplicate(TestsExceptionFilter):
             expected_value="'\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E!,"
         )
 
+    def test_mysql_duplicate_entry_key_start_with_tablename_binary(self):
+        self._run_dupe_constraint_test(
+            "mysql",
+            "(1062, \'Duplicate entry "
+            "\\\'\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E\\\' "
+            "for key \\\'tbl.uniq_tbl0c1\\\'\')",
+            expected_columns=['c1'],
+            expected_value="\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E"
+        )
+        self._run_dupe_constraint_test(
+            "mysql",
+            "(1062, \'Duplicate entry "
+            "''\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E!,' "
+            "for key 'tbl.uniq_tbl0c1'\')",
+            expected_columns=['c1'],
+            expected_value="'\\\\x8A$\\\\x8D\\\\xA6\"s\\\\x8E!,"
+        )
+
     def test_postgresql_single(self):
         self._run_dupe_constraint_test(
             'postgresql',
@@ -604,9 +947,9 @@ class TestDeadlock(TestsExceptionFilter):
         if isinstance(matched, exception.DBError):
             matched = matched.inner_exception
 
-        self.assertEqual(matched.orig.__class__.__name__, expected_dbapi_cls)
+        self.assertEqual(expected_dbapi_cls, matched.orig.__class__.__name__)
 
-    def test_mysql_mysqldb_deadlock(self):
+    def test_mysql_pymysql_deadlock(self):
         self._run_deadlock_detect_test(
             "mysql",
             "(1213, 'Deadlock found when trying "
@@ -614,11 +957,26 @@ class TestDeadlock(TestsExceptionFilter):
             "transaction')"
         )
 
-    def test_mysql_mysqldb_galera_deadlock(self):
+    def test_mysql_pymysql_wsrep_deadlock(self):
+        self._run_deadlock_detect_test(
+            "mysql",
+            "(1213, 'WSREP detected deadlock/conflict and aborted the "
+            "transaction. Try restarting the transaction')",
+            orig_exception_cls=self.InternalError
+        )
+
+        self._run_deadlock_detect_test(
+            "mysql",
+            "(1213, 'Deadlock: wsrep aborted transaction')",
+            orig_exception_cls=self.InternalError
+        )
+
+    def test_mysql_pymysql_galera_deadlock(self):
         self._run_deadlock_detect_test(
             "mysql",
             "(1205, 'Lock wait timeout exceeded; "
-            "try restarting transaction')"
+            "try restarting transaction')",
+            orig_exception_cls=self.InternalError
         )
 
     def test_mysql_mysqlconnector_deadlock(self):
@@ -674,7 +1032,35 @@ class TestDeadlock(TestsExceptionFilter):
         )
 
 
-class IntegrationTest(test_base.DbTestCase):
+class TestDataError(TestsExceptionFilter):
+    def _run_bad_data_test(self, dialect_name, message, error_class):
+        self._run_test(dialect_name,
+                       "INSERT INTO TABLE some_values",
+                       error_class(message),
+                       exception.DBDataError)
+
+    def test_bad_data_incorrect_string(self):
+        # Error sourced from https://bugs.launchpad.net/cinder/+bug/1393871
+        self._run_bad_data_test("mysql",
+                                '(1366, "Incorrect string value: \'\\xF0\' '
+                                'for column \'resource\' at row 1"',
+                                self.OperationalError)
+
+    def test_bad_data_out_of_range(self):
+        # Error sourced from https://bugs.launchpad.net/cinder/+bug/1463379
+        self._run_bad_data_test("mysql",
+                                '(1264, "Out of range value for column '
+                                '\'resource\' at row 1"',
+                                self.DataError)
+
+    def test_data_too_long_for_column(self):
+        self._run_bad_data_test("mysql",
+                                '(1406, "Data too long for column '
+                                '\'resource\' at row 1"',
+                                self.DataError)
+
+
+class IntegrationTest(test_base._DbTestCase):
     """Test an actual error-raising round trips against the database."""
 
     def setUp(self):
@@ -761,7 +1147,7 @@ class IntegrationTest(test_base.DbTestCase):
         foo = self.Foo(counter=sqla.func.imfake(123))
         _session.add(foo)
         matched = self.assertRaises(sqla.exc.OperationalError, _session.flush)
-        self.assertTrue("no such function" in str(matched))
+        self.assertIn("no such function", str(matched))
 
     def test_query_wrapper_operational_error(self):
         """test an operational error from query.all() raised as-is."""
@@ -773,7 +1159,7 @@ class IntegrationTest(test_base.DbTestCase):
         q = _session.query(self.Foo).filter(
             self.Foo.counter == sqla.func.imfake(123))
         matched = self.assertRaises(sqla.exc.OperationalError, q.all)
-        self.assertTrue("no such function" in str(matched))
+        self.assertIn("no such function", str(matched))
 
 
 class TestDBDisconnected(TestsExceptionFilter):
@@ -784,7 +1170,7 @@ class TestDBDisconnected(TestsExceptionFilter):
             dialect_name, exception, num_disconnects, is_disconnect=True):
         engine = self.engine
 
-        event.listen(engine, "engine_connect", session._connect_ping_listener)
+        event.listen(engine, "engine_connect", engines._connect_ping_listener)
 
         real_do_execute = engine.dialect.do_execute
         counter = itertools.count(1)
@@ -811,7 +1197,7 @@ class TestDBDisconnected(TestsExceptionFilter):
         with self._fixture(dialect_name, exc_obj, 1, is_disconnect):
             conn = self.engine.connect()
             with conn.begin():
-                self.assertEqual(conn.scalar(sqla.select([1])), 1)
+                self.assertEqual(1, conn.scalar(sqla.select([1])))
                 self.assertFalse(conn.closed)
                 self.assertFalse(conn.invalidated)
                 self.assertTrue(conn.in_transaction())
@@ -824,7 +1210,23 @@ class TestDBDisconnected(TestsExceptionFilter):
 
         # test implicit execution
         with self._fixture(dialect_name, exc_obj, 1):
-            self.assertEqual(self.engine.scalar(sqla.select([1])), 1)
+            self.assertEqual(1, self.engine.scalar(sqla.select([1])))
+
+    def test_mariadb_error_1927(self):
+        for code in [1927]:
+            self._test_ping_listener_disconnected(
+                "mysql",
+                self.InternalError('%d Connection was killed' % code),
+                is_disconnect=False
+            )
+
+    def test_packet_sequence_wrong_error(self):
+        self._test_ping_listener_disconnected(
+            "mysql",
+            self.InternalError(
+                'Packet sequence number wrong - got 35 expected 1'),
+            is_disconnect=False
+        )
 
     def test_mysql_ping_listener_disconnected(self):
         for code in [2006, 2013, 2014, 2045, 2055]:
@@ -844,6 +1246,24 @@ class TestDBDisconnected(TestsExceptionFilter):
                 is_disconnect=False
             )
 
+    def test_mysql_galera_non_primary_disconnected(self):
+        self._test_ping_listener_disconnected(
+            "mysql",
+            self.OperationalError('(1047, \'Unknown command\') '
+                                  '\'SELECT DATABASE()\' ()')
+        )
+
+    def test_mysql_galera_non_primary_disconnected_regex_only(self):
+        # intentionally set the is_disconnect flag to False
+        # in the "sqlalchemy" layer to make sure the regexp
+        # on _is_db_connection_error is catching
+        self._test_ping_listener_disconnected(
+            "mysql",
+            self.OperationalError('(1047, \'Unknown command\') '
+                                  '\'SELECT DATABASE()\' ()'),
+            is_disconnect=False
+        )
+
     def test_db2_ping_listener_disconnected(self):
         self._test_ping_listener_disconnected(
             "ibm_db_sa",
@@ -895,7 +1315,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
 
         with self._dbapi_fixture(dialect_name):
             with mock.patch.object(engine.dialect, "connect", cant_connect):
-                return session._test_connection(engine, retries, .01)
+                return engines._test_connection(engine, retries, .01)
 
     def test_connect_no_retries(self):
         conn = self._run_test(
@@ -913,7 +1333,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, -1
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_connect_retry_past_failure(self):
         conn = self._run_test(
@@ -922,7 +1342,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, 3
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_connect_retry_not_candidate_exception(self):
         self.assertRaises(
@@ -950,7 +1370,7 @@ class TestDBConnectRetry(TestsExceptionFilter):
             2, -1
         )
         # conn is good
-        self.assertEqual(conn.scalar(sqla.select([1])), 1)
+        self.assertEqual(1, conn.scalar(sqla.select([1])))
 
     def test_db2_error_negative(self):
         self.assertRaises(
@@ -967,7 +1387,7 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
     def setUp(self):
         super(TestDBConnectPingWrapping, self).setUp()
         event.listen(
-            self.engine, "engine_connect", session._connect_ping_listener)
+            self.engine, "engine_connect", engines._connect_ping_listener)
 
     @contextlib.contextmanager
     def _fixture(
@@ -1007,7 +1427,7 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
             self, dialect_name, exc_obj, is_disconnect=True):
         with self._fixture(dialect_name, exc_obj, 3, is_disconnect):
             conn = self.engine.connect()
-            self.assertEqual(conn.scalar(sqla.select([1])), 1)
+            self.assertEqual(1, conn.scalar(sqla.select([1])))
             conn.close()
 
         with self._fixture(dialect_name, exc_obj, 1, is_disconnect):
@@ -1027,15 +1447,15 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
         with self._fixture(dialect_name, exc_obj, 1, is_disconnect):
             self.assertRaises(
                 exception.DBConnectionError,
-                self.engine.contextual_connect
+                self.engine.connect
             )
             self.assertRaises(
                 exception.DBConnectionError,
-                self.engine.contextual_connect
+                self.engine.connect
             )
             self.assertRaises(
                 exception.DBConnectionError,
-                self.engine.contextual_connect
+                self.engine.connect
             )
 
     def test_mysql_w_disconnect_flag(self):
@@ -1052,3 +1472,36 @@ class TestDBConnectPingWrapping(TestsExceptionFilter):
                 self.OperationalError('%d MySQL server has gone away' % code),
                 is_disconnect=False
             )
+
+
+class TestsErrorHandler(TestsExceptionFilter):
+    def test_multiple_error_handlers(self):
+        handler = mock.MagicMock(return_value=None)
+        sqla.event.listen(self.engine, "handle_error", handler, retval=True)
+
+        # cause an error in DB API
+        self._run_test(
+            "mysql", "select you_made_a_programming_error",
+            self.ProgrammingError("Error 123, you made a mistake"),
+            exception.DBError
+        )
+
+        # expect custom handler to be called together with oslo.db's one
+        self.assertEqual(1, handler.call_count,
+                         'Custom handler should be called')
+
+    def test_chained_exceptions(self):
+        class CustomError(Exception):
+            pass
+
+        def handler(context):
+            return CustomError('Custom Error')
+
+        sqla.event.listen(self.engine, "handle_error", handler, retval=True)
+
+        # cause an error in DB API, expect exception from custom handler
+        self._run_test(
+            "mysql", "select you_made_a_programming_error",
+            self.ProgrammingError("Error 123, you made a mistake"),
+            CustomError
+        )
